
/**
 * ISOAK - Iterative similarity optimal assignment kernel.
 * 
 * Written by Matthias Rupp 2006-2007.
 * Copyright (c) 2006-2007, Matthias Rupp, Ewgenij Proschak, Gisbert Schneider.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 *  * The above copyright notice, this permission notice and the following disclaimers
 *    shall be included in all copies or substantial portions of this software.
 *  * The names of the authors may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Please cite
 * 
 * Matthias Rupp, Ewgenij Proschak, Gisbert Schneider:
 * A Kernel Approach to Molecular Similarity Based on Iterative Graph Similarity,
 * Journal of Chemical Information and Molecular Modeling, 47(6): 2280-2286, 2007, 
 * DOI http://dx.doi.org/10.1021/ci700274r.
 *
 * if you use this software.
 */

package info.mrupp.isoak1;

import static info.mrupp.isoak1.Factorial.factorial;
import info.mrupp.isoak1.IVertexEdgeKernel;
import info.mrupp.isoak1.VertexEdgeKernel;
import info.mrupp.isoak1.MolecularGraph;
import info.mrupp.isoak1.Array2Float;
import info.mrupp.isoak1.HungarianAlgorithm;
import info.mrupp.isoak1.HungarianAlgorithm.OptimizationType;
import info.mrupp.isoak1.Permutations;

/**
 * The iterative similarity optimal assignment kernel.
 * 
 * Computes the iterative similarity optimal assignment kernel 
 * between two molecular graphs up to a given precision.
 *
 * <ul>
 * <li>The kernel can be parameterized by the vertex and edge subkernel to be used.</li>
 * <li>If more than one kernel value is to be computed, create one instance of
 *     the IsoaKernel class and reuse it. This will prevent repeated 
 *     reallocation of memory.</li>
 * </ul>
 *   
 * Implementation based on the article<br /><br />
 * 
 * Matthias Rupp, Ewgenij Proschak, Gisbert Schneider:
 * <em>A Kernel Approach to Molecular Similarity Based on Iterative Graph Similarity</em>,
 * Journal of Chemical Information and Molecular Modeling, 2007, submitted.<br /><br />
 * 
 * Please cite this article if you use this software in scientific research.
 * 
 * @author Matthias Rupp
 * @version 1.0 2007-09-17
 *
 */

public class IsoaKernel1 {
    // About the internal data used for computing the kernel (heap
    // storage is retained over instantiation lifetime):
    //
    // On the memory layout of the arrays:
    //
    // Switching from three-dimensional arrays (int [][][]) to two-dimensional
    // arrays (int [][]) improved runtime by 27% on a 175 molecule test set.
    // Switching from two-dimensional arrays (int [][]) to one-dimensional arrays
    // (int []) did not affect runtime.
    // Using a class (Array2Int) decreased performance by 7%. 
    // 
    // Based on these measurements, I decided to use two-dimensional built-in arrays.
    // The memory layout is as follows:
    //
    // The first index equals the index in the similarity vector (the coordinate).
    // For each coordinate, there are up to tau! terms with up to tau summands each.
    // For each summand, there's a factor by which the summand is multiplied and an
    // index (into the similarity vector) which gives the summand itself.
    // Factors and indices are stored separately.
    // The memory layout for the factors (i.j.k denotes the k-th summand factor in
    // the j-th term in the i-th coordinate):
    //
    // 0.0.0, 0.0.1, 0.0.2, ..., 0.0.tau-1,   0.1.0, ..., 0.1.tau-1,   ...,   0.tau!-1.0, ..., 0.tau!-1.tau-1,
    // 1.0.0, 1.0.1, 1.0.2, ..., 1.0.tau-1,   1.1.0, ..., 1.1.tau-1,   ...,   1.tau!-1.0, ..., 1.tau!-1.tau-1,
    // ...
    //
    // For the indices, there's an additional entry for each term giving the number
    // of summands and an additional entry for each coordinate giving the number of terms. 
    // The memory layout for the indices (i.j.k as above, s = number of summands for previous
    // term, t = number of terms):
    //
    // 0.0.0, 0.0.1, 0.0.2, ..., 0.0.tau-1, s,   0.1.0, ..., 0.1.tau-1, s,   ..., 0.tau!-1.0, ..., 0.tau!-1.tau-1, s, t,
    // 1.0.0, 1.0.1, 1.0.2, ..., 1.0.tau-1, s,   1.1.0, ..., 1.1.tau-1, s,   ..., 1.tau!-1.0, ..., 1.tau!-1.tau-1, s, t,    
    // ...
    //
    // Examples (all indices are zero-based):
    //
    // Address of k-th summand factor in j-th term in i-th coordinate: pFactors[i][j*tau+k].
    // Address of k-th summand index in j-th term in i-th coordinate: pIndices[i][j*(tau+1)+k].
    // Number of summands for j-th term in i-th coordinate: pIndices[i][(j+1)*(tau+1)-1].
    // Number of terms in i-th coordinate: pIndices[i][tau!*(tau+1)].
    
    private int kk, nn, kknn;         // Number of atoms in molecules and their product. Class members so there's no need for method parameters.
    private float[] kvec;             // Evaluated vertex kernel, linearized indices.
    private float[][] pFactors;       // Factors for the neighbour assignments. Memory layout as described above.
    private int[][] pIndices;         // Indices for the neighbour assignments. Memory layout as described above.
    private float[] x;                // Similarity matrix, linearized indices.
    private float[] xzero;            // Initial similarity matrix, used for convergence criterion.
    private Array2Float assignTemp;    // Temporary storage for calculation of optimal assignment.
    
    private IVertexEdgeKernel vertexKernel, edgeKernel; // Comparison functions for vertices and edges. 
    private HungarianAlgorithm hungAlg;                 // For computation of optimal assignment.
    
    final static int tau = MolecularGraph.MAX_DEGREE;       // Maximum vertex degree.
    final static int tauOne = tau + 1;                  // Used for indexing pIndices.
    final static int tauFac = factorial(tau);               // tau!.
    final static int pFactorsRow = tauFac*tau;          // Number of elements per coordinate in array pFactors.
    final static int pIndicesRow = tauFac*tauOne + 1;       // Number of elements per coordinate in array pIndices.
    
    /** Initializes internal data structures to an empty state. */
    public IsoaKernel1() {
        kk = 0; nn = 0; kknn = 0;
        kvec = new float[0];
        pFactors = new float[0][0];
        pIndices = new int[0][0];
        x = new float[0];
        xzero = new float[0];
        assignTemp = new Array2Float(0,0);

        vertexKernel = VertexEdgeKernel.NONE.create(new float[]{});
        edgeKernel = VertexEdgeKernel.NONE.create(new float[]{});
        hungAlg = new HungarianAlgorithm();
    }
    
    /** Sets the vertex kernel to be used. Default is NONE. */
    public void setVertexKernel(IVertexEdgeKernel k) { if (k != null) vertexKernel = k; }
    
    /** Sets the edge kernel to be used. Default is NONE. */
    public void setEdgeKernel(IVertexEdgeKernel e) { if (e != null) edgeKernel = e; }
    
    /** 
     * Evaluates the iterative optimal assignment kernel without normalization.
     * 
     * @param molA first argument molecuar graph.
     * @param molB second argument molecular graph.
     * @param alpha the weight of the recursive part of the update equation.
     * @param epsilon the desired precision of the result. The result will not deviate by more than epsilon from the true value, as measured by the maximum-norm based metric.
     * @param assignment  the indices of the optimal assignment. Pass null if not needed.
     * @param similarity  the computed similarity matrix. Pass null if not needed.
     * @return the value of the optimal assignment on the iterative similarity matrix.
     */
    
    public float evalRaw(MolecularGraph molA, MolecularGraph molB, float alpha, float epsilon, int[] assignment, Array2Float similarity) {
        // Set up equations.
        computeInternals(molA, molB);
        int iteration = 0;          // Current number of completed iterations.
        int numIterations = 1;      // Number of iterations that guarantee desired precision. Will be established later on.
        
        if (similarity != null && ( similarity.rows() != kk || similarity.cols() != nn )) throw new RuntimeException(String.format("Size of passed similarity matrix does not match size of passed molecules (expected %d x %d but got %d x %d).", kk, nn, similarity.rows(), similarity.cols()));

        // Check special cases.
        if (kknn == 1) {    // Comparison of two atoms. 
            if (assignment != null) assignment[0] = 0;
            final float result = (kvec[0] + 1)/2; 
            if (similarity != null) similarity.set(0, 0, result);
            return result; 
        }  
        
        // Initialize x to vertex kernel.
        System.arraycopy(kvec, 0, x, 0, kknn);
        
        // Save first value of x for computation of termination criterion.
        System.arraycopy(x, 0, xzero, 0, kknn);
        
        // Iteratively compute new values for x.
        do {
            for (int index = kknn; --index >= 0; ) {
                // Compute the contribution due to neighbour similarity.
                float contrib = 0.f;  
                for (int term = pIndices[index][pIndicesRow-1]; --term >= 0; ) {  // Iterate over max terms.
                    float termContrib = 0.f;
                    for (int summand = pIndices[index][(term+1)*tauOne-1]; --summand >= 0; ) 
                        termContrib += pFactors[index][term*tau+summand] * x[pIndices[index][term*tauOne+summand]];                 
                    contrib = Math.max(contrib, termContrib);
                }
                
                // Update the component.
                x[index] = (1.f-alpha)*kvec[index] + alpha*contrib; // Was: x[index] = (kvec[index] + contrib)/2.f;
            }
            
            // Check termination condition.
            if (++iteration == 1) {  // First iteration, establish number of iterations to do.
                final float distance = maxMetric(x, xzero);
                if (distance <= epsilon*(1.f-alpha)/alpha) numIterations = 1;  // d <= e(1-a)/a <--> e(1-a)/d >= a --> log_{a} (e(1-a)/d) <= 1 --> k >= 1.   
                else numIterations = (int) Math.ceil( Math.log( epsilon*(1.f-alpha)/distance ) / Math.log(alpha) );
            }
        } while (iteration < numIterations);
        
        // Compute optimal assignment.
        for (int row = kk; --row >= 0; ) for(int col = nn; --col >= 0; ) assignTemp.set(row, col, x[col*kk+row]);
        final float result = hungAlg.hungarianAlgorithm(assignTemp, OptimizationType.MAX, assignment);
        
        if (similarity != null) for (int row = kk; --row >= 0; ) for (int col = nn; --col >= 0; ) similarity.set(row, col, x[col*kk+row]);
        
        return result;
    }
    
    /** 
     * Evalutes the iterative optimal assignment kernel with normalization.
     * 
     * The result will be in the range [0,1].
     */
    public float evalNorm(MolecularGraph molA, MolecularGraph molB, float alpha, float epsilon, int[] assignment, Array2Float similarity) {
        float result = evalRaw(molA, molB, alpha, epsilon, assignment, similarity);
        //result /= Math.sqrt(evalRaw(molA, molA, epsilon, null, null) * evalRaw(molB, molB, epsilon, null, null));
        result /= Math.sqrt(molA.numAtoms()*molB.numAtoms());   // Make use of k(G,G) = |V|.
        return result;
    }
    
    /** The maximum-norm based metric is used to compute the number of iterations. */
    private static float maxMetric(float[] x, float[] y) {
        float result = 0.f;
        for (int i = x.length; --i >= 0; ) result = Math.max(result, Math.abs(x[i]-y[i]));
        return result;
    }

    /** Allocates memory and precomputes neighbour assignment matrices. */ 
    private void computeInternals(MolecularGraph molA, MolecularGraph molB) {
        // Shortcuts.
        kk = molA.numAtoms();       // Number of atoms in first molecule.   
        nn = molB.numAtoms();       // Number of atoms in second molecule.
        kknn = kk*nn;               // Number of linearized indices.

        // Allocate memory if necessary.
        if (kvec.length < kknn) kvec = new float[kknn];
        if (pFactors.length < kknn) pFactors = new float[kknn][pFactorsRow];
        if (pIndices.length < kknn) pIndices = new int[kknn][pIndicesRow];
        if (x.length < kknn) x = new float[kknn];           // Only allocated, not initialized.
        if (xzero.length < kknn) xzero = new float[kknn];   // Only allocated, not initialized.
        assignTemp.redim(kk, nn);   // Only allocated, not initialized. Redim is necessary because the Hungarian algorithm expects the right matrix dimensions.
        
        // Compute vertex matchings and P matrices.
        for (int index = 0; index < kknn; ++index) {  // Iterate over linearized entries of similarity vector/matrix.
            // Determine row and column for current index.
            int row = index % kk;
            int col = index / kk;

            // Vertex kernel.
            kvec[index] = vertexKernel.eval(molA, row, molB, col);
            
            // P matrices.
            //
            // To compute the assignment matrices, for each vertex pair, all possible neighbour combinations 
            // are enumerated and stored as terms (at most tau!, each one a neighbourhood assignment) with at most tau summands each.
            // The indices give the involved vertex pairs and the factors include division by degree and the edge kernel.
            final int numNeighbA = molA.numNeighbours(row);
            final int numNeighbB = molB.numNeighbours(col);
            
            if (numNeighbA < numNeighbB) {  // Vertex row has fewer neighbours than vertex col.
                final int[][] permPref = Permutations.prefixesPrecomputed(numNeighbA, numNeighbB);  // Possible assignments.
                pIndices[index][pIndicesRow-1] = permPref.length;                                   // Store number of terms in last entry.
                for (int term = permPref.length; --term >= 0; ) {                                   // Determine each term.
                    pIndices[index][(term+1)*tauOne-1] = numNeighbA;                                // Store number of summands.
                    for (int summand = numNeighbA; --summand >= 0; ) {                              // Determine each summand.
                        final int neighbA = molA.neighbour(row, summand);
                        final int neighbB = molB.neighbour(col, permPref[term][summand]);
                        final int ea = molA.bondIndex(row, neighbA);
                        final int eb = molB.bondIndex(col, neighbB);
                        pFactors[index][term*tau + summand] = (1.f/numNeighbB) * edgeKernel.eval(molA, ea, molB, eb);
                        pIndices[index][term*tauOne + summand] = neighbB*kk + neighbA;
                    }
                }
            } else {  // Vertex col has fewer or equally many neighbours than vertex row. This case is symmetric to the first one.
                final int[][] permPref = Permutations.prefixesPrecomputed(numNeighbB, numNeighbA);
                pIndices[index][pIndicesRow-1] = permPref.length;
                for (int term = permPref.length; --term >= 0; ) {
                    pIndices[index][(term+1)*tauOne-1] = numNeighbB;
                    for (int summand = numNeighbB; --summand >= 0; ) {
                        final int neighbA = molA.neighbour(row, permPref[term][summand]);
                        final int neighbB = molB.neighbour(col, summand);
                        final int ea = molA.bondIndex(row, neighbA);
                        final int eb = molB.bondIndex(col, neighbB);
                        pFactors[index][term*tau + summand] = (1.f/numNeighbA) * edgeKernel.eval(molA, ea, molB, eb);
                        pIndices[index][term*tauOne + summand] = neighbB*kk + neighbA;
                    }                   
                }
            }
        }
    }
}
