
/**
 * ISOAK - Iterative similarity optimal assignment kernel.
 * 
 * Written by Matthias Rupp 2006-2007.
 * Copyright (c) 2006-2007, Matthias Rupp, Ewgenij Proschak, Gisbert Schneider.
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 *  * The above copyright notice, this permission notice and the following disclaimers
 *    shall be included in all copies or substantial portions of this software.
 *  * The names of the authors may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 * Please cite
 * 
 * Matthias Rupp, Ewgenij Proschak, Gisbert Schneider:
 * A Kernel Approach to Molecular Similarity Based on Iterative Graph Similarity,
 * Journal of Chemical Information and Molecular Modeling, 47(6): 2280-2286, 2007, 
 * DOI http://dx.doi.org/10.1021/ci700274r.
 *
 * if you use this software.
 */

package info.mrupp.isoak1;

import static info.mrupp.isoak1.Factorial.factorial;

/**
 * Permutation related functions.
 * 
 * Provides static methods for the computation of all permutations of a given size,
 * all prefixes of given length of permutations of a given size, etc.
 */

public class Permutations {
	/**
	 * All permutations of {0,1,...,n-1}.
	 *
	 * Uses Heaps method.
	 * 
	 * @param n  permutation size. 
	 * @return   all permutations of {0,1,..,n-1} as an array of arrays.
	 */
	public static int[][] permutations(int n) {
		// Special cases.
		if (n < 0) throw new IllegalArgumentException("Permutation size must be non-negative."); 
		if (n == 0) { return new int[0][0]; }
		if (n == 1) { int[][] result = new int[1][1]; result[0][0] = 0; return result; }
		
		// Create result matrix, first row is identity permutation.
		int[][] result = new int[factorial(n)][n];
		for (int i=0; i < n; ++i) result[0][i] = i;
		
		// Initialize indices, buffers etc.
		int pidx = 1;  // Denotes currently created permutation.
		int[] c = new int[n]; for (int i=0; i < n; ++i) c[i] = 1;
		int ii = 1, k;
		
		// Create permutations.
		while (true) {
			if (c[ii] <= ii) {
				if (ii % 2 == 0) { k = 0; } else { k = c[ii]-1; }
				
				System.arraycopy(result[pidx-1], 0, result[pidx], 0, n);  // Copy previous line.
				int t = result[pidx][ii]; result[pidx][ii] = result[pidx][k]; result[pidx][k] = t;  // Swap ii-th and k-th entry.
				++pidx;
				
				c[ii] += 1;
				ii = 1;
			} else {
				c[ii] = 1;
				++ii;
			}
			
			if (ii >= n) break;
		}
		
		return result;
	}
	
	/**
	 * All prefixes of given length of all permutations of given length.
	 *
	 * Generates all length k prefixes of all permutations of {0,1,...,n-1} in lexicographic order.
	 * 
	 * - If k == 0, an empty matrix is returned.
	 * - Once calculated, permutation prefixes are kept across function calls.
	 * - There are (n over k)k! = n!/(n-k)! prefixes of length k of permutations of n elements.
     * 
     * @param k  prefix length.
     * @param n  permutation size.
     * @return   all permutations of {0,1,..,n-1} as an array of arrays.
	 */
	public static int[][] prefixes(int k, int n) {	
		// Special cases.
		if (k < 0 || k > n) throw new IllegalArgumentException("Prefix length must be in range [0,n].");
		if (k == 0) { return new int[0][0]; }
		if (k == n) { return Permutations.permutations(n); }
		
		// Compute prefixes.		
		int[][] result = new int[factorial(n)/factorial(n-k)][k];	// Result matrix.
		for (int i = 0; i < k; ++i) result[0][i] = i;				// Initialize first prefix.
		System.arraycopy(result[0], 0, result[1], 0, k);			// Take over into currently computed result row.
		int residx = 1;												// Current row.
		int idx = k-1;												// Index into current result row, starting at end.
		int[] a = new int[n];										// Symbols not currently in prefix, sorted ascendingly. During computation, additional indices may be temporarily stored.
		for (int i = 0; i < n-k; ++i) a[i] = k+i;					// Initialize remaining symbols.
		int alength = n-k;											// Current size of a (number of occupied entries).
		int j = 0;													// Index into a, starting at beginning.
				
		while (true) {
			// Find first incrementable index from the right.
			while (result[residx][idx] > a[j]) {
				a[alength] = result[residx][idx]; ++alength; 		// Remove large symbol from current prefix and store it in a.
				--idx; ++j;                                     	// Read into s from the right, into a from the left.
				if (idx == -1) return result;	                  	// Only n(n-1)...(n-k+1) can cause this.
			}
			
			// Find smallest number to increment with.
			while (j > 0 && a[j-1] > result[residx][idx]) --j;
			
			// Rebuild suffix.
			int t = result[residx][idx]; result[residx][idx] = a[j]; a[j] = t;  // Swap symbols from s and a.
			if (idx+1 < k) j = 0; else if (j+1 < n-k) ++j;
			while (idx+1 < k) {
				++idx;
				result[residx][idx] = a[0];
				for (int i = 0; i < alength-1; ++i) { a[i] = a[i+1]; } --alength;  // Remove first symbol from a.
			}
			if (residx+1 < result.length) { System.arraycopy(result[residx], 0, result[residx+1], 0, k); ++residx; }
		}
	}
	
	// Precomputed prefixes.
	
	static final int[][][][] prefixPrecomp = {
		{ {} },  // n = 0.
		{ {}, {{0}} },  // n = 1.
		{ {}, {{0},{1}}, {{0,1},{1,0}} },  // n = 2.
		{ {}, {{0},{1},{2}}, {{0,1},{0,2},{1,0},{1,2},{2,0},{2,1}}, {{0,1,2},{0,2,1},{1,0,2},{1,2,0},{2,0,1},{2,1,0}} },  // n = 3.
		{ {}, {{0},{1},{2},{3}}, {{0,1},{0,2},{0,3},{1,0},{1,2},{1,3},{2,0},{2,1},{2,3},{3,0},{3,1},{3,2}},  // n = 4.
			{{0,1,2},{0,1,3},{0,2,1},{0,2,3},{0,3,1},{0,3,2},{1,0,2},{1,0,3},{1,2,0},{1,2,3},{1,3,0},{1,3,2},
			{2,0,1},{2,0,3},{2,1,0},{2,1,3},{2,3,0},{2,3,1},{3,0,1},{3,0,2},{3,1,0},{3,1,2},{3,2,0},{3,2,1}},
			{{0,1,2,3},{0,1,3,2},{0,2,1,3},{0,2,3,1},{0,3,1,2},{0,3,2,1},{1,0,2,3},{1,0,3,2},{1,2,0,3},{1,2,3,0},{1,3,0,2},{1,3,2,0},
			{2,0,1,3},{2,0,3,1},{2,1,0,3},{2,1,3,0},{2,3,0,1},{2,3,1,0},{3,0,1,2},{3,0,2,1},{3,1,0,2},{3,1,2,0},{3,2,0,1},{3,2,1,0}} },
		{ {}, {{0},{1},{2},{3},{4}}, {{0,1},{0,2},{0,3},{0,4},{1,0},{1,2},{1,3},{1,4},{2,0},{2,1},{2,3},{2,4},{3,0},{3,1},{3,2},{3,4},{4,0},{4,1},{4,2},{4,3}},  // n = 5.
			{{0,1,2},{0,1,3},{0,1,4},{0,2,1},{0,2,3},{0,2,4},{0,3,1},{0,3,2},{0,3,4},{0,4,1},{0,4,2},{0,4,3},{1,0,2},{1,0,3},{1,0,4},{1,2,0},{1,2,3},{1,2,4},
			{1,3,0},{1,3,2},{1,3,4},{1,4,0},{1,4,2},{1,4,3},{2,0,1},{2,0,3},{2,0,4},{2,1,0},{2,1,3},{2,1,4},{2,3,0},{2,3,1},{2,3,4},{2,4,0},{2,4,1},{2,4,3},
			{3,0,1},{3,0,2},{3,0,4},{3,1,0},{3,1,2},{3,1,4},{3,2,0},{3,2,1},{3,2,4},{3,4,0},{3,4,1},{3,4,2},{4,0,1},{4,0,2},{4,0,3},{4,1,0},{4,1,2},{4,1,3},
			{4,2,0},{4,2,1},{4,2,3},{4,3,0},{4,3,1},{4,3,2}},
			{ {0,1,2,3},{0,1,2,4},{0,1,3,2},{0,1,3,4},{0,1,4,2},{0,1,4,3},{0,2,1,3},{0,2,1,4},{0,2,3,1},{0,2,3,4},{0,2,4,1},{0,2,4,3},{0,3,1,2},{0,3,1,4},{0,3,2,1},
			{0,3,2,4},{0,3,4,1},{0,3,4,2},{0,4,1,2},{0,4,1,3},{0,4,2,1},{0,4,2,3},{0,4,3,1},{0,4,3,2},{1,0,2,3},{1,0,2,4},{1,0,3,2},{1,0,3,4},{1,0,4,2},{1,0,4,3},
			{1,2,0,3},{1,2,0,4},{1,2,3,0},{1,2,3,4},{1,2,4,0},{1,2,4,3},{1,3,0,2},{1,3,0,4},{1,3,2,0},{1,3,2,4},{1,3,4,0},{1,3,4,2},{1,4,0,2},{1,4,0,3},{1,4,2,0},
			{1,4,2,3},{1,4,3,0},{1,4,3,2},{2,0,1,3},{2,0,1,4},{2,0,3,1},{2,0,3,4},{2,0,4,1},{2,0,4,3},{2,1,0,3},{2,1,0,4},{2,1,3,0},{2,1,3,4},{2,1,4,0},{2,1,4,3},
			{2,3,0,1},{2,3,0,4},{2,3,1,0},{2,3,1,4},{2,3,4,0},{2,3,4,1},{2,4,0,1},{2,4,0,3},{2,4,1,0},{2,4,1,3},{2,4,3,0},{2,4,3,1},{3,0,1,2},{3,0,1,4},{3,0,2,1},
			{3,0,2,4},{3,0,4,1},{3,0,4,2},{3,1,0,2},{3,1,0,4},{3,1,2,0},{3,1,2,4},{3,1,4,0},{3,1,4,2},{3,2,0,1},{3,2,0,4},{3,2,1,0},{3,2,1,4},{3,2,4,0},{3,2,4,1},
			{3,4,0,1},{3,4,0,2},{3,4,1,0},{3,4,1,2},{3,4,2,0},{3,4,2,1},{4,0,1,2},{4,0,1,3},{4,0,2,1},{4,0,2,3},{4,0,3,1},{4,0,3,2},{4,1,0,2},{4,1,0,3},{4,1,2,0},
			{4,1,2,3},{4,1,3,0},{4,1,3,2},{4,2,0,1},{4,2,0,3},{4,2,1,0},{4,2,1,3},{4,2,3,0},{4,2,3,1},{4,3,0,1},{4,3,0,2},{4,3,1,0},{4,3,1,2},{4,3,2,0},{4,3,2,1} },
            { {0,1,2,3,4},{0,1,2,4,3},{0,1,3,2,4},{0,1,3,4,2},{0,1,4,2,3},{0,1,4,3,2},{0,2,1,3,4},{0,2,1,4,3},{0,2,3,1,4},{0,2,3,4,1},{0,2,4,1,3},{0,2,4,3,1},
			{0,3,1,2,4},{0,3,1,4,2},{0,3,2,1,4},{0,3,2,4,1},{0,3,4,1,2},{0,3,4,2,1},{0,4,1,2,3},{0,4,1,3,2},{0,4,2,1,3},{0,4,2,3,1},{0,4,3,1,2},{0,4,3,2,1},{1,0,2,3,4},
			{1,0,2,4,3},{1,0,3,2,4},{1,0,3,4,2},{1,0,4,2,3},{1,0,4,3,2},{1,2,0,3,4},{1,2,0,4,3},{1,2,3,0,4},{1,2,3,4,0},{1,2,4,0,3},{1,2,4,3,0},{1,3,0,2,4},{1,3,0,4,2},
			{1,3,2,0,4},{1,3,2,4,0},{1,3,4,0,2},{1,3,4,2,0},{1,4,0,2,3},{1,4,0,3,2},{1,4,2,0,3},{1,4,2,3,0},{1,4,3,0,2},{1,4,3,2,0},{2,0,1,3,4},{2,0,1,4,3},{2,0,3,1,4},
			{2,0,3,4,1},{2,0,4,1,3},{2,0,4,3,1},{2,1,0,3,4},{2,1,0,4,3},{2,1,3,0,4},{2,1,3,4,0},{2,1,4,0,3},{2,1,4,3,0},{2,3,0,1,4},{2,3,0,4,1},{2,3,1,0,4},{2,3,1,4,0},
			{2,3,4,0,1},{2,3,4,1,0},{2,4,0,1,3},{2,4,0,3,1},{2,4,1,0,3},{2,4,1,3,0},{2,4,3,0,1},{2,4,3,1,0},{3,0,1,2,4},{3,0,1,4,2},{3,0,2,1,4},{3,0,2,4,1},{3,0,4,1,2},
			{3,0,4,2,1},{3,1,0,2,4},{3,1,0,4,2},{3,1,2,0,4},{3,1,2,4,0},{3,1,4,0,2},{3,1,4,2,0},{3,2,0,1,4},{3,2,0,4,1},{3,2,1,0,4},{3,2,1,4,0},{3,2,4,0,1},{3,2,4,1,0},
			{3,4,0,1,2},{3,4,0,2,1},{3,4,1,0,2},{3,4,1,2,0},{3,4,2,0,1},{3,4,2,1,0},{4,0,1,2,3},{4,0,1,3,2},{4,0,2,1,3},{4,0,2,3,1},{4,0,3,1,2},{4,0,3,2,1},{4,1,0,2,3},
			{4,1,0,3,2},{4,1,2,0,3},{4,1,2,3,0},{4,1,3,0,2},{4,1,3,2,0},{4,2,0,1,3},{4,2,0,3,1},{4,2,1,0,3},{4,2,1,3,0},{4,2,3,0,1},{4,2,3,1,0},{4,3,0,1,2},{4,3,0,2,1},
			{4,3,1,0,2},{4,3,1,2,0},{4,3,2,0,1},{4,3,2,1,0} } },
        { {}, { {0},{1},{2},{3},{4},{5} },  // n = 6
          { {0,1},{0,2},{0,3},{0,4},{0,5}, {1,0},{1,2},{1,3},{1,4},{1,5}, {2,0},{2,1},{2,3},{2,4},{2,5},
            {3,0},{3,1},{3,2},{3,4},{3,5}, {4,0},{4,1},{4,2},{4,3},{4,5}, {5,0},{5,1},{5,2},{5,3},{5,4} },
          { {0,1,2},{0,1,3},{0,1,4},{0,1,5},{0,2,1},{0,2,3},{0,2,4},{0,2,5},{0,3,1},{0,3,2},{0,3,4},{0,3,5},{0,4,1},{0,4,2},{0,4,3},{0,4,5},{0,5,1},{0,5,2},{0,5,3},{0,5,4},
            {1,0,2},{1,0,3},{1,0,4},{1,0,5},{1,2,0},{1,2,3},{1,2,4},{1,2,5},{1,3,0},{1,3,2},{1,3,4},{1,3,5},{1,4,0},{1,4,2},{1,4,3},{1,4,5},{1,5,0},{1,5,2},{1,5,3},{1,5,4},
            {2,0,1},{2,0,3},{2,0,4},{2,0,5},{2,1,0},{2,1,3},{2,1,4},{2,1,5},{2,3,0},{2,3,1},{2,3,4},{2,3,5},{2,4,0},{2,4,1},{2,4,3},{2,4,5},{2,5,0},{2,5,1},{2,5,3},{2,5,4},
            {3,0,1},{3,0,2},{3,0,4},{3,0,5},{3,1,0},{3,1,2},{3,1,4},{3,1,5},{3,2,0},{3,2,1},{3,2,4},{3,2,5},{3,4,0},{3,4,1},{3,4,2},{3,4,5},{3,5,0},{3,5,1},{3,5,2},{3,5,4},
            {4,0,1},{4,0,2},{4,0,3},{4,0,5},{4,1,0},{4,1,2},{4,1,3},{4,1,5},{4,2,0},{4,2,1},{4,2,3},{4,2,5},{4,3,0},{4,3,1},{4,3,2},{4,3,5},{4,5,0},{4,5,1},{4,5,2},{4,5,3},
            {5,0,1},{5,0,2},{5,0,3},{5,0,4},{5,1,0},{5,1,2},{5,1,3},{5,1,4},{5,2,0},{5,2,1},{5,2,3},{5,2,4},{5,3,0},{5,3,1},{5,3,2},{5,3,4},{5,4,0},{5,4,1},{5,4,2},{5,4,3} },
          { {0,1,2,3},{0,1,2,4},{0,1,2,5},{0,1,3,2},{0,1,3,4},{0,1,3,5},{0,1,4,2},{0,1,4,3},{0,1,4,5},{0,1,5,2},{0,1,5,3},{0,1,5,4},
            {0,2,1,3},{0,2,1,4},{0,2,1,5},{0,2,3,1},{0,2,3,4},{0,2,3,5},{0,2,4,1},{0,2,4,3},{0,2,4,5},{0,2,5,1},{0,2,5,3},{0,2,5,4},
            {0,3,1,2},{0,3,1,4},{0,3,1,5},{0,3,2,1},{0,3,2,4},{0,3,2,5},{0,3,4,1},{0,3,4,2},{0,3,4,5},{0,3,5,1},{0,3,5,2},{0,3,5,4},
            {0,4,1,2},{0,4,1,3},{0,4,1,5},{0,4,2,1},{0,4,2,3},{0,4,2,5},{0,4,3,1},{0,4,3,2},{0,4,3,5},{0,4,5,1},{0,4,5,2},{0,4,5,3},
            {0,5,1,2},{0,5,1,3},{0,5,1,4},{0,5,2,1},{0,5,2,3},{0,5,2,4},{0,5,3,1},{0,5,3,2},{0,5,3,4},{0,5,4,1},{0,5,4,2},{0,5,4,3},
            {1,0,2,3},{1,0,2,4},{1,0,2,5},{1,0,3,2},{1,0,3,4},{1,0,3,5},{1,0,4,2},{1,0,4,3},{1,0,4,5},{1,0,5,2},{1,0,5,3},{1,0,5,4},
            {1,2,0,3},{1,2,0,4},{1,2,0,5},{1,2,3,0},{1,2,3,4},{1,2,3,5},{1,2,4,0},{1,2,4,3},{1,2,4,5},{1,2,5,0},{1,2,5,3},{1,2,5,4},
            {1,3,0,2},{1,3,0,4},{1,3,0,5},{1,3,2,0},{1,3,2,4},{1,3,2,5},{1,3,4,0},{1,3,4,2},{1,3,4,5},{1,3,5,0},{1,3,5,2},{1,3,5,4},
            {1,4,0,2},{1,4,0,3},{1,4,0,5},{1,4,2,0},{1,4,2,3},{1,4,2,5},{1,4,3,0},{1,4,3,2},{1,4,3,5},{1,4,5,0},{1,4,5,2},{1,4,5,3},
            {1,5,0,2},{1,5,0,3},{1,5,0,4},{1,5,2,0},{1,5,2,3},{1,5,2,4},{1,5,3,0},{1,5,3,2},{1,5,3,4},{1,5,4,0},{1,5,4,2},{1,5,4,3},
            {2,0,1,3},{2,0,1,4},{2,0,1,5},{2,0,3,1},{2,0,3,4},{2,0,3,5},{2,0,4,1},{2,0,4,3},{2,0,4,5},{2,0,5,1},{2,0,5,3},{2,0,5,4},
            {2,1,0,3},{2,1,0,4},{2,1,0,5},{2,1,3,0},{2,1,3,4},{2,1,3,5},{2,1,4,0},{2,1,4,3},{2,1,4,5},{2,1,5,0},{2,1,5,3},{2,1,5,4},
            {2,3,0,1},{2,3,0,4},{2,3,0,5},{2,3,1,0},{2,3,1,4},{2,3,1,5},{2,3,4,0},{2,3,4,1},{2,3,4,5},{2,3,5,0},{2,3,5,1},{2,3,5,4},
            {2,4,0,1},{2,4,0,3},{2,4,0,5},{2,4,1,0},{2,4,1,3},{2,4,1,5},{2,4,3,0},{2,4,3,1},{2,4,3,5},{2,4,5,0},{2,4,5,1},{2,4,5,3},
            {2,5,0,1},{2,5,0,3},{2,5,0,4},{2,5,1,0},{2,5,1,3},{2,5,1,4},{2,5,3,0},{2,5,3,1},{2,5,3,4},{2,5,4,0},{2,5,4,1},{2,5,4,3},
            {3,0,1,2},{3,0,1,4},{3,0,1,5},{3,0,2,1},{3,0,2,4},{3,0,2,5},{3,0,4,1},{3,0,4,2},{3,0,4,5},{3,0,5,1},{3,0,5,2},{3,0,5,4},
            {3,1,0,2},{3,1,0,4},{3,1,0,5},{3,1,2,0},{3,1,2,4},{3,1,2,5},{3,1,4,0},{3,1,4,2},{3,1,4,5},{3,1,5,0},{3,1,5,2},{3,1,5,4},
            {3,2,0,1},{3,2,0,4},{3,2,0,5},{3,2,1,0},{3,2,1,4},{3,2,1,5},{3,2,4,0},{3,2,4,1},{3,2,4,5},{3,2,5,0},{3,2,5,1},{3,2,5,4},
            {3,4,0,1},{3,4,0,2},{3,4,0,5},{3,4,1,0},{3,4,1,2},{3,4,1,5},{3,4,2,0},{3,4,2,1},{3,4,2,5},{3,4,5,0},{3,4,5,1},{3,4,5,2},
            {3,5,0,1},{3,5,0,2},{3,5,0,4},{3,5,1,0},{3,5,1,2},{3,5,1,4},{3,5,2,0},{3,5,2,1},{3,5,2,4},{3,5,4,0},{3,5,4,1},{3,5,4,2},
            {4,0,1,2},{4,0,1,3},{4,0,1,5},{4,0,2,1},{4,0,2,3},{4,0,2,5},{4,0,3,1},{4,0,3,2},{4,0,3,5},{4,0,5,1},{4,0,5,2},{4,0,5,3},
            {4,1,0,2},{4,1,0,3},{4,1,0,5},{4,1,2,0},{4,1,2,3},{4,1,2,5},{4,1,3,0},{4,1,3,2},{4,1,3,5},{4,1,5,0},{4,1,5,2},{4,1,5,3},
            {4,2,0,1},{4,2,0,3},{4,2,0,5},{4,2,1,0},{4,2,1,3},{4,2,1,5},{4,2,3,0},{4,2,3,1},{4,2,3,5},{4,2,5,0},{4,2,5,1},{4,2,5,3},
            {4,3,0,1},{4,3,0,2},{4,3,0,5},{4,3,1,0},{4,3,1,2},{4,3,1,5},{4,3,2,0},{4,3,2,1},{4,3,2,5},{4,3,5,0},{4,3,5,1},{4,3,5,2},
            {4,5,0,1},{4,5,0,2},{4,5,0,3},{4,5,1,0},{4,5,1,2},{4,5,1,3},{4,5,2,0},{4,5,2,1},{4,5,2,3},{4,5,3,0},{4,5,3,1},{4,5,3,2},
            {5,0,1,2},{5,0,1,3},{5,0,1,4},{5,0,2,1},{5,0,2,3},{5,0,2,4},{5,0,3,1},{5,0,3,2},{5,0,3,4},{5,0,4,1},{5,0,4,2},{5,0,4,3},
            {5,1,0,2},{5,1,0,3},{5,1,0,4},{5,1,2,0},{5,1,2,3},{5,1,2,4},{5,1,3,0},{5,1,3,2},{5,1,3,4},{5,1,4,0},{5,1,4,2},{5,1,4,3},
            {5,2,0,1},{5,2,0,3},{5,2,0,4},{5,2,1,0},{5,2,1,3},{5,2,1,4},{5,2,3,0},{5,2,3,1},{5,2,3,4},{5,2,4,0},{5,2,4,1},{5,2,4,3},
            {5,3,0,1},{5,3,0,2},{5,3,0,4},{5,3,1,0},{5,3,1,2},{5,3,1,4},{5,3,2,0},{5,3,2,1},{5,3,2,4},{5,3,4,0},{5,3,4,1},{5,3,4,2},
            {5,4,0,1},{5,4,0,2},{5,4,0,3},{5,4,1,0},{5,4,1,2},{5,4,1,3},{5,4,2,0},{5,4,2,1},{5,4,2,3},{5,4,3,0},{5,4,3,1},{5,4,3,2} },             
          { {0,1,2,3,4},{0,1,2,3,5},{0,1,2,4,3},{0,1,2,4,5},{0,1,2,5,3},{0,1,2,5,4},{0,1,3,2,4},{0,1,3,2,5},{0,1,3,4,2},{0,1,3,4,5},{0,1,3,5,2},{0,1,3,5,4},
            {0,1,4,2,3},{0,1,4,2,5},{0,1,4,3,2},{0,1,4,3,5},{0,1,4,5,2},{0,1,4,5,3},{0,1,5,2,3},{0,1,5,2,4},{0,1,5,3,2},{0,1,5,3,4},{0,1,5,4,2},{0,1,5,4,3},
            {0,2,1,3,4},{0,2,1,3,5},{0,2,1,4,3},{0,2,1,4,5},{0,2,1,5,3},{0,2,1,5,4},{0,2,3,1,4},{0,2,3,1,5},{0,2,3,4,1},{0,2,3,4,5},{0,2,3,5,1},{0,2,3,5,4},
            {0,2,4,1,3},{0,2,4,1,5},{0,2,4,3,1},{0,2,4,3,5},{0,2,4,5,1},{0,2,4,5,3},{0,2,5,1,3},{0,2,5,1,4},{0,2,5,3,1},{0,2,5,3,4},{0,2,5,4,1},{0,2,5,4,3},
            {0,3,1,2,4},{0,3,1,2,5},{0,3,1,4,2},{0,3,1,4,5},{0,3,1,5,2},{0,3,1,5,4},{0,3,2,1,4},{0,3,2,1,5},{0,3,2,4,1},{0,3,2,4,5},{0,3,2,5,1},{0,3,2,5,4},
            {0,3,4,1,2},{0,3,4,1,5},{0,3,4,2,1},{0,3,4,2,5},{0,3,4,5,1},{0,3,4,5,2},{0,3,5,1,2},{0,3,5,1,4},{0,3,5,2,1},{0,3,5,2,4},{0,3,5,4,1},{0,3,5,4,2},
            {0,4,1,2,3},{0,4,1,2,5},{0,4,1,3,2},{0,4,1,3,5},{0,4,1,5,2},{0,4,1,5,3},{0,4,2,1,3},{0,4,2,1,5},{0,4,2,3,1},{0,4,2,3,5},{0,4,2,5,1},{0,4,2,5,3},
            {0,4,3,1,2},{0,4,3,1,5},{0,4,3,2,1},{0,4,3,2,5},{0,4,3,5,1},{0,4,3,5,2},{0,4,5,1,2},{0,4,5,1,3},{0,4,5,2,1},{0,4,5,2,3},{0,4,5,3,1},{0,4,5,3,2},
            {0,5,1,2,3},{0,5,1,2,4},{0,5,1,3,2},{0,5,1,3,4},{0,5,1,4,2},{0,5,1,4,3},{0,5,2,1,3},{0,5,2,1,4},{0,5,2,3,1},{0,5,2,3,4},{0,5,2,4,1},{0,5,2,4,3},
            {0,5,3,1,2},{0,5,3,1,4},{0,5,3,2,1},{0,5,3,2,4},{0,5,3,4,1},{0,5,3,4,2},{0,5,4,1,2},{0,5,4,1,3},{0,5,4,2,1},{0,5,4,2,3},{0,5,4,3,1},{0,5,4,3,2},
            {1,0,2,3,4},{1,0,2,3,5},{1,0,2,4,3},{1,0,2,4,5},{1,0,2,5,3},{1,0,2,5,4},{1,0,3,2,4},{1,0,3,2,5},{1,0,3,4,2},{1,0,3,4,5},{1,0,3,5,2},{1,0,3,5,4},
            {1,0,4,2,3},{1,0,4,2,5},{1,0,4,3,2},{1,0,4,3,5},{1,0,4,5,2},{1,0,4,5,3},{1,0,5,2,3},{1,0,5,2,4},{1,0,5,3,2},{1,0,5,3,4},{1,0,5,4,2},{1,0,5,4,3},
            {1,2,0,3,4},{1,2,0,3,5},{1,2,0,4,3},{1,2,0,4,5},{1,2,0,5,3},{1,2,0,5,4},{1,2,3,0,4},{1,2,3,0,5},{1,2,3,4,0},{1,2,3,4,5},{1,2,3,5,0},{1,2,3,5,4},
            {1,2,4,0,3},{1,2,4,0,5},{1,2,4,3,0},{1,2,4,3,5},{1,2,4,5,0},{1,2,4,5,3},{1,2,5,0,3},{1,2,5,0,4},{1,2,5,3,0},{1,2,5,3,4},{1,2,5,4,0},{1,2,5,4,3},
            {1,3,0,2,4},{1,3,0,2,5},{1,3,0,4,2},{1,3,0,4,5},{1,3,0,5,2},{1,3,0,5,4},{1,3,2,0,4},{1,3,2,0,5},{1,3,2,4,0},{1,3,2,4,5},{1,3,2,5,0},{1,3,2,5,4},
            {1,3,4,0,2},{1,3,4,0,5},{1,3,4,2,0},{1,3,4,2,5},{1,3,4,5,0},{1,3,4,5,2},{1,3,5,0,2},{1,3,5,0,4},{1,3,5,2,0},{1,3,5,2,4},{1,3,5,4,0},{1,3,5,4,2},
            {1,4,0,2,3},{1,4,0,2,5},{1,4,0,3,2},{1,4,0,3,5},{1,4,0,5,2},{1,4,0,5,3},{1,4,2,0,3},{1,4,2,0,5},{1,4,2,3,0},{1,4,2,3,5},{1,4,2,5,0},{1,4,2,5,3},
            {1,4,3,0,2},{1,4,3,0,5},{1,4,3,2,0},{1,4,3,2,5},{1,4,3,5,0},{1,4,3,5,2},{1,4,5,0,2},{1,4,5,0,3},{1,4,5,2,0},{1,4,5,2,3},{1,4,5,3,0},{1,4,5,3,2},
            {1,5,0,2,3},{1,5,0,2,4},{1,5,0,3,2},{1,5,0,3,4},{1,5,0,4,2},{1,5,0,4,3},{1,5,2,0,3},{1,5,2,0,4},{1,5,2,3,0},{1,5,2,3,4},{1,5,2,4,0},{1,5,2,4,3},
            {1,5,3,0,2},{1,5,3,0,4},{1,5,3,2,0},{1,5,3,2,4},{1,5,3,4,0},{1,5,3,4,2},{1,5,4,0,2},{1,5,4,0,3},{1,5,4,2,0},{1,5,4,2,3},{1,5,4,3,0},{1,5,4,3,2},
            {2,0,1,3,4},{2,0,1,3,5},{2,0,1,4,3},{2,0,1,4,5},{2,0,1,5,3},{2,0,1,5,4},{2,0,3,1,4},{2,0,3,1,5},{2,0,3,4,1},{2,0,3,4,5},{2,0,3,5,1},{2,0,3,5,4},
            {2,0,4,1,3},{2,0,4,1,5},{2,0,4,3,1},{2,0,4,3,5},{2,0,4,5,1},{2,0,4,5,3},{2,0,5,1,3},{2,0,5,1,4},{2,0,5,3,1},{2,0,5,3,4},{2,0,5,4,1},{2,0,5,4,3},
            {2,1,0,3,4},{2,1,0,3,5},{2,1,0,4,3},{2,1,0,4,5},{2,1,0,5,3},{2,1,0,5,4},{2,1,3,0,4},{2,1,3,0,5},{2,1,3,4,0},{2,1,3,4,5},{2,1,3,5,0},{2,1,3,5,4},
            {2,1,4,0,3},{2,1,4,0,5},{2,1,4,3,0},{2,1,4,3,5},{2,1,4,5,0},{2,1,4,5,3},{2,1,5,0,3},{2,1,5,0,4},{2,1,5,3,0},{2,1,5,3,4},{2,1,5,4,0},{2,1,5,4,3},
            {2,3,0,1,4},{2,3,0,1,5},{2,3,0,4,1},{2,3,0,4,5},{2,3,0,5,1},{2,3,0,5,4},{2,3,1,0,4},{2,3,1,0,5},{2,3,1,4,0},{2,3,1,4,5},{2,3,1,5,0},{2,3,1,5,4},
            {2,3,4,0,1},{2,3,4,0,5},{2,3,4,1,0},{2,3,4,1,5},{2,3,4,5,0},{2,3,4,5,1},{2,3,5,0,1},{2,3,5,0,4},{2,3,5,1,0},{2,3,5,1,4},{2,3,5,4,0},{2,3,5,4,1},
            {2,4,0,1,3},{2,4,0,1,5},{2,4,0,3,1},{2,4,0,3,5},{2,4,0,5,1},{2,4,0,5,3},{2,4,1,0,3},{2,4,1,0,5},{2,4,1,3,0},{2,4,1,3,5},{2,4,1,5,0},{2,4,1,5,3},
            {2,4,3,0,1},{2,4,3,0,5},{2,4,3,1,0},{2,4,3,1,5},{2,4,3,5,0},{2,4,3,5,1},{2,4,5,0,1},{2,4,5,0,3},{2,4,5,1,0},{2,4,5,1,3},{2,4,5,3,0},{2,4,5,3,1},
            {2,5,0,1,3},{2,5,0,1,4},{2,5,0,3,1},{2,5,0,3,4},{2,5,0,4,1},{2,5,0,4,3},{2,5,1,0,3},{2,5,1,0,4},{2,5,1,3,0},{2,5,1,3,4},{2,5,1,4,0},{2,5,1,4,3},
            {2,5,3,0,1},{2,5,3,0,4},{2,5,3,1,0},{2,5,3,1,4},{2,5,3,4,0},{2,5,3,4,1},{2,5,4,0,1},{2,5,4,0,3},{2,5,4,1,0},{2,5,4,1,3},{2,5,4,3,0},{2,5,4,3,1},
            {3,0,1,2,4},{3,0,1,2,5},{3,0,1,4,2},{3,0,1,4,5},{3,0,1,5,2},{3,0,1,5,4},{3,0,2,1,4},{3,0,2,1,5},{3,0,2,4,1},{3,0,2,4,5},{3,0,2,5,1},{3,0,2,5,4},
            {3,0,4,1,2},{3,0,4,1,5},{3,0,4,2,1},{3,0,4,2,5},{3,0,4,5,1},{3,0,4,5,2},{3,0,5,1,2},{3,0,5,1,4},{3,0,5,2,1},{3,0,5,2,4},{3,0,5,4,1},{3,0,5,4,2},
            {3,1,0,2,4},{3,1,0,2,5},{3,1,0,4,2},{3,1,0,4,5},{3,1,0,5,2},{3,1,0,5,4},{3,1,2,0,4},{3,1,2,0,5},{3,1,2,4,0},{3,1,2,4,5},{3,1,2,5,0},{3,1,2,5,4},
            {3,1,4,0,2},{3,1,4,0,5},{3,1,4,2,0},{3,1,4,2,5},{3,1,4,5,0},{3,1,4,5,2},{3,1,5,0,2},{3,1,5,0,4},{3,1,5,2,0},{3,1,5,2,4},{3,1,5,4,0},{3,1,5,4,2},
            {3,2,0,1,4},{3,2,0,1,5},{3,2,0,4,1},{3,2,0,4,5},{3,2,0,5,1},{3,2,0,5,4},{3,2,1,0,4},{3,2,1,0,5},{3,2,1,4,0},{3,2,1,4,5},{3,2,1,5,0},{3,2,1,5,4},
            {3,2,4,0,1},{3,2,4,0,5},{3,2,4,1,0},{3,2,4,1,5},{3,2,4,5,0},{3,2,4,5,1},{3,2,5,0,1},{3,2,5,0,4},{3,2,5,1,0},{3,2,5,1,4},{3,2,5,4,0},{3,2,5,4,1},
            {3,4,0,1,2},{3,4,0,1,5},{3,4,0,2,1},{3,4,0,2,5},{3,4,0,5,1},{3,4,0,5,2},{3,4,1,0,2},{3,4,1,0,5},{3,4,1,2,0},{3,4,1,2,5},{3,4,1,5,0},{3,4,1,5,2},
            {3,4,2,0,1},{3,4,2,0,5},{3,4,2,1,0},{3,4,2,1,5},{3,4,2,5,0},{3,4,2,5,1},{3,4,5,0,1},{3,4,5,0,2},{3,4,5,1,0},{3,4,5,1,2},{3,4,5,2,0},{3,4,5,2,1},
            {3,5,0,1,2},{3,5,0,1,4},{3,5,0,2,1},{3,5,0,2,4},{3,5,0,4,1},{3,5,0,4,2},{3,5,1,0,2},{3,5,1,0,4},{3,5,1,2,0},{3,5,1,2,4},{3,5,1,4,0},{3,5,1,4,2},
            {3,5,2,0,1},{3,5,2,0,4},{3,5,2,1,0},{3,5,2,1,4},{3,5,2,4,0},{3,5,2,4,1},{3,5,4,0,1},{3,5,4,0,2},{3,5,4,1,0},{3,5,4,1,2},{3,5,4,2,0},{3,5,4,2,1},
            {4,0,1,2,3},{4,0,1,2,5},{4,0,1,3,2},{4,0,1,3,5},{4,0,1,5,2},{4,0,1,5,3},{4,0,2,1,3},{4,0,2,1,5},{4,0,2,3,1},{4,0,2,3,5},{4,0,2,5,1},{4,0,2,5,3},
            {4,0,3,1,2},{4,0,3,1,5},{4,0,3,2,1},{4,0,3,2,5},{4,0,3,5,1},{4,0,3,5,2},{4,0,5,1,2},{4,0,5,1,3},{4,0,5,2,1},{4,0,5,2,3},{4,0,5,3,1},{4,0,5,3,2},
            {4,1,0,2,3},{4,1,0,2,5},{4,1,0,3,2},{4,1,0,3,5},{4,1,0,5,2},{4,1,0,5,3},{4,1,2,0,3},{4,1,2,0,5},{4,1,2,3,0},{4,1,2,3,5},{4,1,2,5,0},{4,1,2,5,3},
            {4,1,3,0,2},{4,1,3,0,5},{4,1,3,2,0},{4,1,3,2,5},{4,1,3,5,0},{4,1,3,5,2},{4,1,5,0,2},{4,1,5,0,3},{4,1,5,2,0},{4,1,5,2,3},{4,1,5,3,0},{4,1,5,3,2},
            {4,2,0,1,3},{4,2,0,1,5},{4,2,0,3,1},{4,2,0,3,5},{4,2,0,5,1},{4,2,0,5,3},{4,2,1,0,3},{4,2,1,0,5},{4,2,1,3,0},{4,2,1,3,5},{4,2,1,5,0},{4,2,1,5,3},
            {4,2,3,0,1},{4,2,3,0,5},{4,2,3,1,0},{4,2,3,1,5},{4,2,3,5,0},{4,2,3,5,1},{4,2,5,0,1},{4,2,5,0,3},{4,2,5,1,0},{4,2,5,1,3},{4,2,5,3,0},{4,2,5,3,1},
            {4,3,0,1,2},{4,3,0,1,5},{4,3,0,2,1},{4,3,0,2,5},{4,3,0,5,1},{4,3,0,5,2},{4,3,1,0,2},{4,3,1,0,5},{4,3,1,2,0},{4,3,1,2,5},{4,3,1,5,0},{4,3,1,5,2},
            {4,3,2,0,1},{4,3,2,0,5},{4,3,2,1,0},{4,3,2,1,5},{4,3,2,5,0},{4,3,2,5,1},{4,3,5,0,1},{4,3,5,0,2},{4,3,5,1,0},{4,3,5,1,2},{4,3,5,2,0},{4,3,5,2,1},
            {4,5,0,1,2},{4,5,0,1,3},{4,5,0,2,1},{4,5,0,2,3},{4,5,0,3,1},{4,5,0,3,2},{4,5,1,0,2},{4,5,1,0,3},{4,5,1,2,0},{4,5,1,2,3},{4,5,1,3,0},{4,5,1,3,2},
            {4,5,2,0,1},{4,5,2,0,3},{4,5,2,1,0},{4,5,2,1,3},{4,5,2,3,0},{4,5,2,3,1},{4,5,3,0,1},{4,5,3,0,2},{4,5,3,1,0},{4,5,3,1,2},{4,5,3,2,0},{4,5,3,2,1},
            {5,0,1,2,3},{5,0,1,2,4},{5,0,1,3,2},{5,0,1,3,4},{5,0,1,4,2},{5,0,1,4,3},{5,0,2,1,3},{5,0,2,1,4},{5,0,2,3,1},{5,0,2,3,4},{5,0,2,4,1},{5,0,2,4,3},
            {5,0,3,1,2},{5,0,3,1,4},{5,0,3,2,1},{5,0,3,2,4},{5,0,3,4,1},{5,0,3,4,2},{5,0,4,1,2},{5,0,4,1,3},{5,0,4,2,1},{5,0,4,2,3},{5,0,4,3,1},{5,0,4,3,2},
            {5,1,0,2,3},{5,1,0,2,4},{5,1,0,3,2},{5,1,0,3,4},{5,1,0,4,2},{5,1,0,4,3},{5,1,2,0,3},{5,1,2,0,4},{5,1,2,3,0},{5,1,2,3,4},{5,1,2,4,0},{5,1,2,4,3},
            {5,1,3,0,2},{5,1,3,0,4},{5,1,3,2,0},{5,1,3,2,4},{5,1,3,4,0},{5,1,3,4,2},{5,1,4,0,2},{5,1,4,0,3},{5,1,4,2,0},{5,1,4,2,3},{5,1,4,3,0},{5,1,4,3,2},
            {5,2,0,1,3},{5,2,0,1,4},{5,2,0,3,1},{5,2,0,3,4},{5,2,0,4,1},{5,2,0,4,3},{5,2,1,0,3},{5,2,1,0,4},{5,2,1,3,0},{5,2,1,3,4},{5,2,1,4,0},{5,2,1,4,3},
            {5,2,3,0,1},{5,2,3,0,4},{5,2,3,1,0},{5,2,3,1,4},{5,2,3,4,0},{5,2,3,4,1},{5,2,4,0,1},{5,2,4,0,3},{5,2,4,1,0},{5,2,4,1,3},{5,2,4,3,0},{5,2,4,3,1},
            {5,3,0,1,2},{5,3,0,1,4},{5,3,0,2,1},{5,3,0,2,4},{5,3,0,4,1},{5,3,0,4,2},{5,3,1,0,2},{5,3,1,0,4},{5,3,1,2,0},{5,3,1,2,4},{5,3,1,4,0},{5,3,1,4,2},
            {5,3,2,0,1},{5,3,2,0,4},{5,3,2,1,0},{5,3,2,1,4},{5,3,2,4,0},{5,3,2,4,1},{5,3,4,0,1},{5,3,4,0,2},{5,3,4,1,0},{5,3,4,1,2},{5,3,4,2,0},{5,3,4,2,1},
            {5,4,0,1,2},{5,4,0,1,3},{5,4,0,2,1},{5,4,0,2,3},{5,4,0,3,1},{5,4,0,3,2},{5,4,1,0,2},{5,4,1,0,3},{5,4,1,2,0},{5,4,1,2,3},{5,4,1,3,0},{5,4,1,3,2},
            {5,4,2,0,1},{5,4,2,0,3},{5,4,2,1,0},{5,4,2,1,3},{5,4,2,3,0},{5,4,2,3,1},{5,4,3,0,1},{5,4,3,0,2},{5,4,3,1,0},{5,4,3,1,2},{5,4,3,2,0},{5,4,3,2,1} },
          { {0,1,2,3,4,5},{0,1,2,3,5,4},{0,1,2,4,3,5},{0,1,2,4,5,3},{0,1,2,5,3,4},{0,1,2,5,4,3},{0,1,3,2,4,5},{0,1,3,2,5,4},{0,1,3,4,2,5},{0,1,3,4,5,2},
            {0,1,3,5,2,4},{0,1,3,5,4,2},{0,1,4,2,3,5},{0,1,4,2,5,3},{0,1,4,3,2,5},{0,1,4,3,5,2},{0,1,4,5,2,3},{0,1,4,5,3,2},{0,1,5,2,3,4},{0,1,5,2,4,3},
            {0,1,5,3,2,4},{0,1,5,3,4,2},{0,1,5,4,2,3},{0,1,5,4,3,2},{0,2,1,3,4,5},{0,2,1,3,5,4},{0,2,1,4,3,5},{0,2,1,4,5,3},{0,2,1,5,3,4},{0,2,1,5,4,3},
            {0,2,3,1,4,5},{0,2,3,1,5,4},{0,2,3,4,1,5},{0,2,3,4,5,1},{0,2,3,5,1,4},{0,2,3,5,4,1},{0,2,4,1,3,5},{0,2,4,1,5,3},{0,2,4,3,1,5},{0,2,4,3,5,1},
            {0,2,4,5,1,3},{0,2,4,5,3,1},{0,2,5,1,3,4},{0,2,5,1,4,3},{0,2,5,3,1,4},{0,2,5,3,4,1},{0,2,5,4,1,3},{0,2,5,4,3,1},{0,3,1,2,4,5},{0,3,1,2,5,4},
            {0,3,1,4,2,5},{0,3,1,4,5,2},{0,3,1,5,2,4},{0,3,1,5,4,2},{0,3,2,1,4,5},{0,3,2,1,5,4},{0,3,2,4,1,5},{0,3,2,4,5,1},{0,3,2,5,1,4},{0,3,2,5,4,1},
            {0,3,4,1,2,5},{0,3,4,1,5,2},{0,3,4,2,1,5},{0,3,4,2,5,1},{0,3,4,5,1,2},{0,3,4,5,2,1},{0,3,5,1,2,4},{0,3,5,1,4,2},{0,3,5,2,1,4},{0,3,5,2,4,1},
            {0,3,5,4,1,2},{0,3,5,4,2,1},{0,4,1,2,3,5},{0,4,1,2,5,3},{0,4,1,3,2,5},{0,4,1,3,5,2},{0,4,1,5,2,3},{0,4,1,5,3,2},{0,4,2,1,3,5},{0,4,2,1,5,3},
            {0,4,2,3,1,5},{0,4,2,3,5,1},{0,4,2,5,1,3},{0,4,2,5,3,1},{0,4,3,1,2,5},{0,4,3,1,5,2},{0,4,3,2,1,5},{0,4,3,2,5,1},{0,4,3,5,1,2},{0,4,3,5,2,1},
            {0,4,5,1,2,3},{0,4,5,1,3,2},{0,4,5,2,1,3},{0,4,5,2,3,1},{0,4,5,3,1,2},{0,4,5,3,2,1},{0,5,1,2,3,4},{0,5,1,2,4,3},{0,5,1,3,2,4},{0,5,1,3,4,2},
            {0,5,1,4,2,3},{0,5,1,4,3,2},{0,5,2,1,3,4},{0,5,2,1,4,3},{0,5,2,3,1,4},{0,5,2,3,4,1},{0,5,2,4,1,3},{0,5,2,4,3,1},{0,5,3,1,2,4},{0,5,3,1,4,2},
            {0,5,3,2,1,4},{0,5,3,2,4,1},{0,5,3,4,1,2},{0,5,3,4,2,1},{0,5,4,1,2,3},{0,5,4,1,3,2},{0,5,4,2,1,3},{0,5,4,2,3,1},{0,5,4,3,1,2},{0,5,4,3,2,1},
            {1,0,2,3,4,5},{1,0,2,3,5,4},{1,0,2,4,3,5},{1,0,2,4,5,3},{1,0,2,5,3,4},{1,0,2,5,4,3},{1,0,3,2,4,5},{1,0,3,2,5,4},{1,0,3,4,2,5},{1,0,3,4,5,2},
            {1,0,3,5,2,4},{1,0,3,5,4,2},{1,0,4,2,3,5},{1,0,4,2,5,3},{1,0,4,3,2,5},{1,0,4,3,5,2},{1,0,4,5,2,3},{1,0,4,5,3,2},{1,0,5,2,3,4},{1,0,5,2,4,3},
            {1,0,5,3,2,4},{1,0,5,3,4,2},{1,0,5,4,2,3},{1,0,5,4,3,2},{1,2,0,3,4,5},{1,2,0,3,5,4},{1,2,0,4,3,5},{1,2,0,4,5,3},{1,2,0,5,3,4},{1,2,0,5,4,3},
            {1,2,3,0,4,5},{1,2,3,0,5,4},{1,2,3,4,0,5},{1,2,3,4,5,0},{1,2,3,5,0,4},{1,2,3,5,4,0},{1,2,4,0,3,5},{1,2,4,0,5,3},{1,2,4,3,0,5},{1,2,4,3,5,0},
            {1,2,4,5,0,3},{1,2,4,5,3,0},{1,2,5,0,3,4},{1,2,5,0,4,3},{1,2,5,3,0,4},{1,2,5,3,4,0},{1,2,5,4,0,3},{1,2,5,4,3,0},{1,3,0,2,4,5},{1,3,0,2,5,4},
            {1,3,0,4,2,5},{1,3,0,4,5,2},{1,3,0,5,2,4},{1,3,0,5,4,2},{1,3,2,0,4,5},{1,3,2,0,5,4},{1,3,2,4,0,5},{1,3,2,4,5,0},{1,3,2,5,0,4},{1,3,2,5,4,0},
            {1,3,4,0,2,5},{1,3,4,0,5,2},{1,3,4,2,0,5},{1,3,4,2,5,0},{1,3,4,5,0,2},{1,3,4,5,2,0},{1,3,5,0,2,4},{1,3,5,0,4,2},{1,3,5,2,0,4},{1,3,5,2,4,0},
            {1,3,5,4,0,2},{1,3,5,4,2,0},{1,4,0,2,3,5},{1,4,0,2,5,3},{1,4,0,3,2,5},{1,4,0,3,5,2},{1,4,0,5,2,3},{1,4,0,5,3,2},{1,4,2,0,3,5},{1,4,2,0,5,3},
            {1,4,2,3,0,5},{1,4,2,3,5,0},{1,4,2,5,0,3},{1,4,2,5,3,0},{1,4,3,0,2,5},{1,4,3,0,5,2},{1,4,3,2,0,5},{1,4,3,2,5,0},{1,4,3,5,0,2},{1,4,3,5,2,0},
            {1,4,5,0,2,3},{1,4,5,0,3,2},{1,4,5,2,0,3},{1,4,5,2,3,0},{1,4,5,3,0,2},{1,4,5,3,2,0},{1,5,0,2,3,4},{1,5,0,2,4,3},{1,5,0,3,2,4},{1,5,0,3,4,2},
            {1,5,0,4,2,3},{1,5,0,4,3,2},{1,5,2,0,3,4},{1,5,2,0,4,3},{1,5,2,3,0,4},{1,5,2,3,4,0},{1,5,2,4,0,3},{1,5,2,4,3,0},{1,5,3,0,2,4},{1,5,3,0,4,2},
            {1,5,3,2,0,4},{1,5,3,2,4,0},{1,5,3,4,0,2},{1,5,3,4,2,0},{1,5,4,0,2,3},{1,5,4,0,3,2},{1,5,4,2,0,3},{1,5,4,2,3,0},{1,5,4,3,0,2},{1,5,4,3,2,0},
            {2,0,1,3,4,5},{2,0,1,3,5,4},{2,0,1,4,3,5},{2,0,1,4,5,3},{2,0,1,5,3,4},{2,0,1,5,4,3},{2,0,3,1,4,5},{2,0,3,1,5,4},{2,0,3,4,1,5},{2,0,3,4,5,1},
            {2,0,3,5,1,4},{2,0,3,5,4,1},{2,0,4,1,3,5},{2,0,4,1,5,3},{2,0,4,3,1,5},{2,0,4,3,5,1},{2,0,4,5,1,3},{2,0,4,5,3,1},{2,0,5,1,3,4},{2,0,5,1,4,3},
            {2,0,5,3,1,4},{2,0,5,3,4,1},{2,0,5,4,1,3},{2,0,5,4,3,1},{2,1,0,3,4,5},{2,1,0,3,5,4},{2,1,0,4,3,5},{2,1,0,4,5,3},{2,1,0,5,3,4},{2,1,0,5,4,3},
            {2,1,3,0,4,5},{2,1,3,0,5,4},{2,1,3,4,0,5},{2,1,3,4,5,0},{2,1,3,5,0,4},{2,1,3,5,4,0},{2,1,4,0,3,5},{2,1,4,0,5,3},{2,1,4,3,0,5},{2,1,4,3,5,0},
            {2,1,4,5,0,3},{2,1,4,5,3,0},{2,1,5,0,3,4},{2,1,5,0,4,3},{2,1,5,3,0,4},{2,1,5,3,4,0},{2,1,5,4,0,3},{2,1,5,4,3,0},{2,3,0,1,4,5},{2,3,0,1,5,4},
            {2,3,0,4,1,5},{2,3,0,4,5,1},{2,3,0,5,1,4},{2,3,0,5,4,1},{2,3,1,0,4,5},{2,3,1,0,5,4},{2,3,1,4,0,5},{2,3,1,4,5,0},{2,3,1,5,0,4},{2,3,1,5,4,0},
            {2,3,4,0,1,5},{2,3,4,0,5,1},{2,3,4,1,0,5},{2,3,4,1,5,0},{2,3,4,5,0,1},{2,3,4,5,1,0},{2,3,5,0,1,4},{2,3,5,0,4,1},{2,3,5,1,0,4},{2,3,5,1,4,0},
            {2,3,5,4,0,1},{2,3,5,4,1,0},{2,4,0,1,3,5},{2,4,0,1,5,3},{2,4,0,3,1,5},{2,4,0,3,5,1},{2,4,0,5,1,3},{2,4,0,5,3,1},{2,4,1,0,3,5},{2,4,1,0,5,3},
            {2,4,1,3,0,5},{2,4,1,3,5,0},{2,4,1,5,0,3},{2,4,1,5,3,0},{2,4,3,0,1,5},{2,4,3,0,5,1},{2,4,3,1,0,5},{2,4,3,1,5,0},{2,4,3,5,0,1},{2,4,3,5,1,0},
            {2,4,5,0,1,3},{2,4,5,0,3,1},{2,4,5,1,0,3},{2,4,5,1,3,0},{2,4,5,3,0,1},{2,4,5,3,1,0},{2,5,0,1,3,4},{2,5,0,1,4,3},{2,5,0,3,1,4},{2,5,0,3,4,1},
            {2,5,0,4,1,3},{2,5,0,4,3,1},{2,5,1,0,3,4},{2,5,1,0,4,3},{2,5,1,3,0,4},{2,5,1,3,4,0},{2,5,1,4,0,3},{2,5,1,4,3,0},{2,5,3,0,1,4},{2,5,3,0,4,1},
            {2,5,3,1,0,4},{2,5,3,1,4,0},{2,5,3,4,0,1},{2,5,3,4,1,0},{2,5,4,0,1,3},{2,5,4,0,3,1},{2,5,4,1,0,3},{2,5,4,1,3,0},{2,5,4,3,0,1},{2,5,4,3,1,0},
            {3,0,1,2,4,5},{3,0,1,2,5,4},{3,0,1,4,2,5},{3,0,1,4,5,2},{3,0,1,5,2,4},{3,0,1,5,4,2},{3,0,2,1,4,5},{3,0,2,1,5,4},{3,0,2,4,1,5},{3,0,2,4,5,1},
            {3,0,2,5,1,4},{3,0,2,5,4,1},{3,0,4,1,2,5},{3,0,4,1,5,2},{3,0,4,2,1,5},{3,0,4,2,5,1},{3,0,4,5,1,2},{3,0,4,5,2,1},{3,0,5,1,2,4},{3,0,5,1,4,2},
            {3,0,5,2,1,4},{3,0,5,2,4,1},{3,0,5,4,1,2},{3,0,5,4,2,1},{3,1,0,2,4,5},{3,1,0,2,5,4},{3,1,0,4,2,5},{3,1,0,4,5,2},{3,1,0,5,2,4},{3,1,0,5,4,2},
            {3,1,2,0,4,5},{3,1,2,0,5,4},{3,1,2,4,0,5},{3,1,2,4,5,0},{3,1,2,5,0,4},{3,1,2,5,4,0},{3,1,4,0,2,5},{3,1,4,0,5,2},{3,1,4,2,0,5},{3,1,4,2,5,0},
            {3,1,4,5,0,2},{3,1,4,5,2,0},{3,1,5,0,2,4},{3,1,5,0,4,2},{3,1,5,2,0,4},{3,1,5,2,4,0},{3,1,5,4,0,2},{3,1,5,4,2,0},{3,2,0,1,4,5},{3,2,0,1,5,4},
            {3,2,0,4,1,5},{3,2,0,4,5,1},{3,2,0,5,1,4},{3,2,0,5,4,1},{3,2,1,0,4,5},{3,2,1,0,5,4},{3,2,1,4,0,5},{3,2,1,4,5,0},{3,2,1,5,0,4},{3,2,1,5,4,0},
            {3,2,4,0,1,5},{3,2,4,0,5,1},{3,2,4,1,0,5},{3,2,4,1,5,0},{3,2,4,5,0,1},{3,2,4,5,1,0},{3,2,5,0,1,4},{3,2,5,0,4,1},{3,2,5,1,0,4},{3,2,5,1,4,0},
            {3,2,5,4,0,1},{3,2,5,4,1,0},{3,4,0,1,2,5},{3,4,0,1,5,2},{3,4,0,2,1,5},{3,4,0,2,5,1},{3,4,0,5,1,2},{3,4,0,5,2,1},{3,4,1,0,2,5},{3,4,1,0,5,2},
            {3,4,1,2,0,5},{3,4,1,2,5,0},{3,4,1,5,0,2},{3,4,1,5,2,0},{3,4,2,0,1,5},{3,4,2,0,5,1},{3,4,2,1,0,5},{3,4,2,1,5,0},{3,4,2,5,0,1},{3,4,2,5,1,0},
            {3,4,5,0,1,2},{3,4,5,0,2,1},{3,4,5,1,0,2},{3,4,5,1,2,0},{3,4,5,2,0,1},{3,4,5,2,1,0},{3,5,0,1,2,4},{3,5,0,1,4,2},{3,5,0,2,1,4},{3,5,0,2,4,1},
            {3,5,0,4,1,2},{3,5,0,4,2,1},{3,5,1,0,2,4},{3,5,1,0,4,2},{3,5,1,2,0,4},{3,5,1,2,4,0},{3,5,1,4,0,2},{3,5,1,4,2,0},{3,5,2,0,1,4},{3,5,2,0,4,1},
            {3,5,2,1,0,4},{3,5,2,1,4,0},{3,5,2,4,0,1},{3,5,2,4,1,0},{3,5,4,0,1,2},{3,5,4,0,2,1},{3,5,4,1,0,2},{3,5,4,1,2,0},{3,5,4,2,0,1},{3,5,4,2,1,0},
            {4,0,1,2,3,5},{4,0,1,2,5,3},{4,0,1,3,2,5},{4,0,1,3,5,2},{4,0,1,5,2,3},{4,0,1,5,3,2},{4,0,2,1,3,5},{4,0,2,1,5,3},{4,0,2,3,1,5},{4,0,2,3,5,1},
            {4,0,2,5,1,3},{4,0,2,5,3,1},{4,0,3,1,2,5},{4,0,3,1,5,2},{4,0,3,2,1,5},{4,0,3,2,5,1},{4,0,3,5,1,2},{4,0,3,5,2,1},{4,0,5,1,2,3},{4,0,5,1,3,2},
            {4,0,5,2,1,3},{4,0,5,2,3,1},{4,0,5,3,1,2},{4,0,5,3,2,1},{4,1,0,2,3,5},{4,1,0,2,5,3},{4,1,0,3,2,5},{4,1,0,3,5,2},{4,1,0,5,2,3},{4,1,0,5,3,2},
            {4,1,2,0,3,5},{4,1,2,0,5,3},{4,1,2,3,0,5},{4,1,2,3,5,0},{4,1,2,5,0,3},{4,1,2,5,3,0},{4,1,3,0,2,5},{4,1,3,0,5,2},{4,1,3,2,0,5},{4,1,3,2,5,0},
            {4,1,3,5,0,2},{4,1,3,5,2,0},{4,1,5,0,2,3},{4,1,5,0,3,2},{4,1,5,2,0,3},{4,1,5,2,3,0},{4,1,5,3,0,2},{4,1,5,3,2,0},{4,2,0,1,3,5},{4,2,0,1,5,3},
            {4,2,0,3,1,5},{4,2,0,3,5,1},{4,2,0,5,1,3},{4,2,0,5,3,1},{4,2,1,0,3,5},{4,2,1,0,5,3},{4,2,1,3,0,5},{4,2,1,3,5,0},{4,2,1,5,0,3},{4,2,1,5,3,0},
            {4,2,3,0,1,5},{4,2,3,0,5,1},{4,2,3,1,0,5},{4,2,3,1,5,0},{4,2,3,5,0,1},{4,2,3,5,1,0},{4,2,5,0,1,3},{4,2,5,0,3,1},{4,2,5,1,0,3},{4,2,5,1,3,0},
            {4,2,5,3,0,1},{4,2,5,3,1,0},{4,3,0,1,2,5},{4,3,0,1,5,2},{4,3,0,2,1,5},{4,3,0,2,5,1},{4,3,0,5,1,2},{4,3,0,5,2,1},{4,3,1,0,2,5},{4,3,1,0,5,2},
            {4,3,1,2,0,5},{4,3,1,2,5,0},{4,3,1,5,0,2},{4,3,1,5,2,0},{4,3,2,0,1,5},{4,3,2,0,5,1},{4,3,2,1,0,5},{4,3,2,1,5,0},{4,3,2,5,0,1},{4,3,2,5,1,0},
            {4,3,5,0,1,2},{4,3,5,0,2,1},{4,3,5,1,0,2},{4,3,5,1,2,0},{4,3,5,2,0,1},{4,3,5,2,1,0},{4,5,0,1,2,3},{4,5,0,1,3,2},{4,5,0,2,1,3},{4,5,0,2,3,1},
            {4,5,0,3,1,2},{4,5,0,3,2,1},{4,5,1,0,2,3},{4,5,1,0,3,2},{4,5,1,2,0,3},{4,5,1,2,3,0},{4,5,1,3,0,2},{4,5,1,3,2,0},{4,5,2,0,1,3},{4,5,2,0,3,1},
            {4,5,2,1,0,3},{4,5,2,1,3,0},{4,5,2,3,0,1},{4,5,2,3,1,0},{4,5,3,0,1,2},{4,5,3,0,2,1},{4,5,3,1,0,2},{4,5,3,1,2,0},{4,5,3,2,0,1},{4,5,3,2,1,0},
            {5,0,1,2,3,4},{5,0,1,2,4,3},{5,0,1,3,2,4},{5,0,1,3,4,2},{5,0,1,4,2,3},{5,0,1,4,3,2},{5,0,2,1,3,4},{5,0,2,1,4,3},{5,0,2,3,1,4},{5,0,2,3,4,1},
            {5,0,2,4,1,3},{5,0,2,4,3,1},{5,0,3,1,2,4},{5,0,3,1,4,2},{5,0,3,2,1,4},{5,0,3,2,4,1},{5,0,3,4,1,2},{5,0,3,4,2,1},{5,0,4,1,2,3},{5,0,4,1,3,2},
            {5,0,4,2,1,3},{5,0,4,2,3,1},{5,0,4,3,1,2},{5,0,4,3,2,1},{5,1,0,2,3,4},{5,1,0,2,4,3},{5,1,0,3,2,4},{5,1,0,3,4,2},{5,1,0,4,2,3},{5,1,0,4,3,2},
            {5,1,2,0,3,4},{5,1,2,0,4,3},{5,1,2,3,0,4},{5,1,2,3,4,0},{5,1,2,4,0,3},{5,1,2,4,3,0},{5,1,3,0,2,4},{5,1,3,0,4,2},{5,1,3,2,0,4},{5,1,3,2,4,0},
            {5,1,3,4,0,2},{5,1,3,4,2,0},{5,1,4,0,2,3},{5,1,4,0,3,2},{5,1,4,2,0,3},{5,1,4,2,3,0},{5,1,4,3,0,2},{5,1,4,3,2,0},{5,2,0,1,3,4},{5,2,0,1,4,3},
            {5,2,0,3,1,4},{5,2,0,3,4,1},{5,2,0,4,1,3},{5,2,0,4,3,1},{5,2,1,0,3,4},{5,2,1,0,4,3},{5,2,1,3,0,4},{5,2,1,3,4,0},{5,2,1,4,0,3},{5,2,1,4,3,0},
            {5,2,3,0,1,4},{5,2,3,0,4,1},{5,2,3,1,0,4},{5,2,3,1,4,0},{5,2,3,4,0,1},{5,2,3,4,1,0},{5,2,4,0,1,3},{5,2,4,0,3,1},{5,2,4,1,0,3},{5,2,4,1,3,0},
            {5,2,4,3,0,1},{5,2,4,3,1,0},{5,3,0,1,2,4},{5,3,0,1,4,2},{5,3,0,2,1,4},{5,3,0,2,4,1},{5,3,0,4,1,2},{5,3,0,4,2,1},{5,3,1,0,2,4},{5,3,1,0,4,2},
            {5,3,1,2,0,4},{5,3,1,2,4,0},{5,3,1,4,0,2},{5,3,1,4,2,0},{5,3,2,0,1,4},{5,3,2,0,4,1},{5,3,2,1,0,4},{5,3,2,1,4,0},{5,3,2,4,0,1},{5,3,2,4,1,0},
            {5,3,4,0,1,2},{5,3,4,0,2,1},{5,3,4,1,0,2},{5,3,4,1,2,0},{5,3,4,2,0,1},{5,3,4,2,1,0},{5,4,0,1,2,3},{5,4,0,1,3,2},{5,4,0,2,1,3},{5,4,0,2,3,1},
            {5,4,0,3,1,2},{5,4,0,3,2,1},{5,4,1,0,2,3},{5,4,1,0,3,2},{5,4,1,2,0,3},{5,4,1,2,3,0},{5,4,1,3,0,2},{5,4,1,3,2,0},{5,4,2,0,1,3},{5,4,2,0,3,1},
            {5,4,2,1,0,3},{5,4,2,1,3,0},{5,4,2,3,0,1},{5,4,2,3,1,0},{5,4,3,0,1,2},{5,4,3,0,2,1},{5,4,3,1,0,2},{5,4,3,1,2,0},{5,4,3,2,0,1},{5,4,3,2,1,0} }            
        }
	};

    /**
     * Precomputed prefixes of given length of permutations of given length.
     *
     * Generates all length k prefixes of all permutations of {0,1,...,n-1} in lexicographic order.
     * Currently, 0 <= n <= 6.
     * 
     * - If k == 0, an empty matrix is returned.
     * - Once calculated, permutation prefixes are kept across function calls.
     * - There are (n over k)k! = n!/(n-k)! prefixes of length k of permutations of n elements.
     * 
     * @param k  prefix length.
     * @param n  permutation size.
     * @return   all permutations of {0,1,..,n-1} as an array of arrays.
     */
	public static int[][] prefixesPrecomputed(int k, int n) {
		try { return prefixPrecomp[n][k]; } 
		catch(ArrayIndexOutOfBoundsException e) { throw new IllegalArgumentException(String.format("Unsupported argument (%s) for precomputed prefixes", e.getMessage()));	}
	}
    
    /** All functionality is provided via static methods, so there is no need to construct objects of this type. */
    public Permutations() {}
}
